@@ -43,7 +43,6 @@ NeqStore: TypeAlias = list[list[tuple[Var, Value]]]
 DomainStore: TypeAlias = PMap[Var, set[int]]
 ConstraintFunction: TypeAlias = Callable[["State"], Optional["State"]]
 ConstraintStore: TypeAlias = list["Constraint"]
-StreamThunk: TypeAlias = Callable[[], "Stream"]
 
 
 def empty_sub() -> Substitution:
@@ -103,7 +102,7 @@ Stream: TypeAlias = tuple[()] | Callable[[], "Stream"] | tuple[State, "Stream"]
 
 
 class GoalProto(Protocol):
-    def __call__(self, state: State) -> StreamThunk:
+    def __call__(self, state: State) -> Stream:
         ...
 
 
@@ -117,7 +116,7 @@ class Goal:
         update_wrapper(self, goal)
         self.goal = goal
 
-    def __call__(self, state: State) -> StreamThunk:
+    def __call__(self, state: State) -> Stream:
         return lambda: self.goal(state)
 
     def __or__(self, other):
@@ -129,7 +128,7 @@ class Goal:
 
 
 def goal_from_constraint(constraint: ConstraintFunction) -> GoalProto:
-    def _goal(state: State) -> StreamThunk:
+    def _goal(state: State) -> Stream:
         match constraint(state):
             case None:
                 return mzero
@@ -150,7 +149,7 @@ def unit(state: State) -> Stream:
     return (state, mzero)
 
 
-def mplus(s1: Stream, s2: Stream) -> StreamThunk:
+def mplus(s1: Stream, s2: Stream) -> Stream:
     match s1:
         case ():
             return lambda: s2
@@ -162,7 +161,7 @@ def mplus(s1: Stream, s2: Stream) -> StreamThunk:
             raise InvalidStream
 
 
-def bind(stream: Stream, g: GoalProto) -> StreamThunk:
+def bind(stream: Stream, g: GoalProto) -> Stream:
     match stream:
         case ():
             return mzero
@@ -248,7 +247,7 @@ def fail() -> GoalProto:
 
 
 def snooze(g: GoalConstructorProto, *args: Value) -> GoalProto:
-    def delayed_goal(state) -> StreamThunk:
+    def delayed_goal(state) -> Stream:
         return g(*args)(state)
 
     return Goal(delayed_goal)
@@ -261,12 +260,11 @@ def delay(g: GoalProto) -> GoalProto:
 def disj(g: GoalProto, *goals: GoalProto) -> GoalProto:
     if goals == ():
         return delay(g)
-    g2, *rest = goals
-    return _disj(delay(g), disj(g2, *rest))
+    return reduce(_disj, (delay(goal) for goal in goals), delay(g))
 
 
 def _disj(g1: GoalProto, g2: GoalProto) -> GoalProto:
-    def __disj(state: State) -> StreamThunk:
+    def __disj(state: State) -> Stream:
         return mplus(g1(state), g2(state))
 
     return Goal(__disj)
@@ -275,12 +273,11 @@ def _disj(g1: GoalProto, g2: GoalProto) -> GoalProto:
 def conj(g: GoalProto, *goals: GoalProto) -> GoalProto:
     if goals == ():
         return delay(g)
-    g2, *rest = goals
-    return _conj(delay(g), conj(g2, *rest))
+    return reduce(_conj, (delay(goal) for goal in goals), delay(g))
 
 
 def _conj(g1: GoalProto, g2: GoalProto) -> GoalProto:
-    def __conj(state: State) -> StreamThunk:
+    def __conj(state: State) -> Stream:
         return bind(g1(state), g2)
 
     return Goal(__conj)
@@ -297,7 +294,9 @@ def eq(u: Value, v: Value) -> GoalProto:
             return state
         else:
             prefix = get_sub_prefix(new_sub, state.sub)
-            return process_prefix(prefix, state.constraints)(state.set(sub=new_sub))
+            return Hooks.process_prefix(prefix, state.constraints)(
+                state.set(sub=new_sub)
+            )
 
     return goal_from_constraint(_eqc)
 
@@ -668,7 +667,7 @@ def process_prefix_fd(
 
 
 def enforce_constraints_fd(x: Var) -> GoalProto:
-    def _enforce_constraints(state: State) -> StreamThunk:
+    def _enforce_constraints(state: State) -> Stream:
         bound_vars = state.domains.keys()
         verify_all_bound(state.constraints, bound_vars)
         return onceo(force_answer(bound_vars))(state)
@@ -706,7 +705,7 @@ def verify_all_bound(constraints: ConstraintStore, bound_vars: list[Var]):
 
 
 def force_answer(x: Var | list[Var]) -> GoalProto:
-    def _force_answer(state: State) -> StreamThunk:
+    def _force_answer(state: State) -> Stream:
         match walk(x, state.sub):
             case Var(_) as var if (d := state.get_domain(var)) is not None:
                 return map_sum(lambda val: eq(x, val), d)(state)
@@ -756,9 +755,9 @@ def starfoldr(f, xs, initial):
 
 
 def ifte(g1, g2, g3=fail()) -> GoalProto:
-    def _ifte(state: State) -> StreamThunk:
+    def _ifte(state: State) -> Stream:
         # TODO: rewrite iteratively
-        def ifte_loop(stream: Stream) -> StreamThunk:
+        def ifte_loop(stream: Stream) -> Stream:
             match stream:
                 case ():
                     return g3(state)
@@ -858,10 +857,7 @@ def reify_state(state: State, v: Var) -> Value:
     v = walk_all(v, state.sub)
     reified_sub = reify_sub(v, empty_sub())
     v = walk_all(v, reified_sub)
-    return v
-    # if len(state.constraints) == 0:
-    #     return v
-    # return reify_constraints(v, reified_sub)(state)
+    return Hooks.reify_value(Hooks.reify_constraints(v, reified_sub, state))
 
 
 def walk_all(v: Value, s: Substitution) -> Value:
@@ -884,7 +880,7 @@ def reify_sub(v: Value, s: Substitution) -> Substitution:
     v = walk(v, s)
     match v:
         case _ if isinstance(v, Var):
-            return extend_substitution(v, ReifiedVar(len(s)), s)
+            return extend_substitution(v, Hooks.reify_var(v, s), s)
         case (a, *d) | cons(a, d):
             return reify_sub(d, reify_sub(a, s))
         case _:
@@ -901,7 +897,7 @@ def run(n: int, f_fresh_vars: Callable[..., GoalProto]):
     state = State.empty().set(var_count=n_vars)
     goal = conj(
         f_fresh_vars(*fresh_vars),
-        *map(enforce_constraints, fresh_vars),
+        *map(Hooks.enforce_constraints, fresh_vars),
     )
     return reify(take(n, goal(state)), *fresh_vars)
 
@@ -912,7 +908,7 @@ def run_all(f_fresh_vars: Callable[..., GoalProto]):
     state = State.empty().set(var_count=n_vars)
     goal = conj(
         f_fresh_vars(*fresh_vars),
-        *map(enforce_constraints, fresh_vars),
+        *map(Hooks.enforce_constraints, fresh_vars),
     )
     return reify(take_all(goal(state)), *fresh_vars)
 
@@ -923,7 +919,7 @@ def irun(f_fresh_vars: Callable[..., GoalProto]):
     state = State.empty().set(var_count=n_vars)
     goal = conj(
         f_fresh_vars(*fresh_vars),
-        *map(enforce_constraints, fresh_vars),
+        *map(Hooks.enforce_constraints, fresh_vars),
     )
     return ireify(itake(goal(state)), *fresh_vars)
 
@@ -936,48 +932,42 @@ def default_enforce_constraints(*_):
     return succeed()
 
 
-def default_reify_constraints(*_):
-    return succeed()
+def default_reify_constraints(value, *_):
+    return value
 
 
-__PROCESS_PREFIX__: Callable[
-    [Substitution, ConstraintStore],
-    ConstraintFunction,
-] = default_process_prefix
-__ENFORCE_CONSTRAINTS__: Callable[[Var], GoalProto] = default_enforce_constraints
-__REIFY_CONSTRAINTS__: Callable[
-    [Var, Substitution],
-    GoalProto,
-] = default_reify_constraints
+def default_reify_var(_, substitution):
+    return ReifiedVar(len(substitution))
 
 
-def set_process_prefix(constraint_function):
-    global __PROCESS_PREFIX__
-    __PROCESS_PREFIX__ = constraint_function
+def default_reify_value(value):
+    return value
 
 
-def set_enforce_constraints(goal):
-    global __ENFORCE_CONSTRAINTS__
-    __ENFORCE_CONSTRAINTS__ = goal
+class HooksMeta(type):
+    def _set_hook(cls, name):
+        def update_hooks(hook_function):
+            if not hasattr(cls, name):
+                raise AttributeError(
+                    f"Cannot set unknown hook '{name}' on registry '{cls.__name__}'"
+                )
+            setattr(cls, name, hook_function)
+
+        return update_hooks
+
+    def __getattribute__(cls, attr):
+        if attr.startswith("set_"):
+            return cls._set_hook(attr[4:])
+        return super().__getattribute__(attr)
 
 
-def set_reify_constraints(goal):
-    global __REIFY_CONSTRAINTS__
-    __REIFY_CONSTRAINTS__ = goal
-
-
-def process_prefix(
-    prefix: Substitution, constraints: ConstraintStore
-) -> ConstraintFunction:
-    global __PROCESS_PREFIX__
-    return __PROCESS_PREFIX__(prefix, constraints)
-
-
-def enforce_constraints(x: Var) -> GoalProto:
-    global __ENFORCE_CONSTRAINTS__
-    return __ENFORCE_CONSTRAINTS__(x)
-
-
-def reify_constraints(x: Var, s: Substitution) -> GoalProto:
-    global __REIFY_CONSTRAINTS__
-    return __REIFY_CONSTRAINTS__(x, s)
+class Hooks(metaclass=HooksMeta):
+    process_prefix: Callable[
+        [Substitution, ConstraintStore], ConstraintFunction
+    ] = default_process_prefix
+    enforce_constraints: Callable[[Var], GoalProto] = default_enforce_constraints
+    reify_constraints: Callable[
+        [Value, Substitution, State], Any
+    ] = default_reify_constraints
+    reify_var: Callable[[Var, Substitution], Any] = default_reify_var
+    reify_value: Callable[[Any], Any] = default_reify_value
